import hashlib, os


def md5sum(file):
    md5sum = hashlib.md5()
    with open(file, "rb") as f:
        while chunk := f.read():
            md5sum.update(chunk)
    return md5sum


def upload_one(path):
    from pyncm import GetCurrentSession
    from pyncm.apis.cloud import (
        GetCheckCloudUpload,
        GetNosToken,
        SetUploadObject,
        SetUploadCloudInfo,
        SetPublishCloudResource,
    )

    fname = os.path.basename(path)
    fext = path.split(".")[-1]
    """Parsing file names"""
    fsize = os.stat(path).st_size
    md5 = md5sum(path).hexdigest()
    print("MD5", md5)
    cresult = GetCheckCloudUpload(md5)
    songId = cresult["songId"]
    """网盘资源发布 4 步走：
    1.拿到上传令牌 - 需要文件名，MD5，文件大小"""
    token = GetNosToken(fname, md5, fsize, fext)["result"]
    if cresult["needUpload"]:
        print("开始上传 %s ( %s B )" % (fname, fsize))
        """2. 若文件未曾上传完毕，则完成其上传"""
        upload_result = SetUploadObject(
            open(path, "rb"), md5, fsize, token["objectKey"], token["token"]
        )
        print("响应", upload_result)
    print(
        f"""准备提交
    ID  :   {songId}
    MD5 :   {md5}
    NAME:   {fname}"""
    )
    submit_result = SetUploadCloudInfo(
        resourceId=token["resourceId"],
        songid=songId,
        md5=md5,
        filename=fname,
        song=fname,
        artist=GetCurrentSession().nickname,
        album="PyNCM",
        bitrate=1000,
    )
    """3. 提交资源"""
    print("提交响应", submit_result)
    """4. 发布资源"""
    publish_result = SetPublishCloudResource(submit_result["songId"])
    print("发布结果", publish_result)


if __name__ == "__main__":
    from __init__ import login

    assert login(), "登录失败"
    # login via phone,may change as you like
    upload_one(input("文件路径："))



# -*- coding: utf-8 -*-
"""视频 - Video - 有关 APIs"""
from . import WeapiCryptoRequest


@WeapiCryptoRequest
def GetMVDetail(mv_id: str):
    """网页端 - 获取 MV 详情

    Args:
        mv_id (str): MV ID

    Returns:
        dict
    """
    return "/weapi/v1/mv/detail", {"id": str(mv_id)}


@WeapiCryptoRequest
def GetMVResource(mv_id: str, res=1080):
    """网页端 - 获取 MV 音视频资源

    Args:
        mv_id (str): MV ID
        res (int, optional): 240 / 480 / 720 / 1080. Defaults to 1080.

    Returns:
        dict
    """
    return "/weapi/song/enhance/play/mv/url", {"id": str(mv_id), "r": str(res)}


@WeapiCryptoRequest
def GetMVComments(mv_id: str, offset=0, limit=20, total=False):
    """网页端 - 获取MV评论

    Args:
        mv_id (str): MV ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    """
    return "/weapi/v1/resource/comments/R_MV_5_%s" % mv_id, {
        "rid": "R_MV_5_%s" % mv_id,
        "offset": str(offset),
        "total": str(total).lower(),
        "limit": str(limit),
    }



from __init__ import login
from pyncm.apis.track import GetTrackAudio

assert login(), "登录失败"
print(GetTrackAudio(input("歌曲 ID:"), bitrate=3200 * 1000))
"""一般地 320k即以上即 flac, 320k及以下即 mp3,96k及以下即 m4a
取决于歌曲；若 bitrate 值非常大，则回传最佳音质歌曲（如本例）
"""



# -*- coding: utf-8 -*-
"""歌曲 - Track APIs"""
from . import EapiCryptoRequest, EapiEncipered, WeapiCryptoRequest, LoginRequiredApi
from ..utils import RandomString
import json


@WeapiCryptoRequest
def GetTrackDetail(song_ids: list):
    """网页端 - 获取歌曲详情

    Args:
        song_ids (list): 歌曲 ID

    Notes:
        song_ids 项目数应 <= 1000

    Returns:
        dict
    """
    ids = song_ids if isinstance(song_ids, list) else [song_ids]
    return "/weapi/v3/song/detail", {"c": json.dumps([{"id": str(id)} for id in ids])}


@EapiCryptoRequest
def GetTrackAudio(song_ids: list, bitrate=320000, encodeType="aac"):
    """PC 端 - 获取歌曲音频详情（文件URL、MD5...）

    Args:
        song_ids (list): 歌曲 ID
        bitrate (int, optional): 比特率 (96k SQ 320k HQ 320k+ Lossless/SQ). Defaults to 320000
        encodeType (str, optional) Defaults to `aac`. 使用高 bitrate 值时无视该选项

    Notes:
        song_ids 项目数应 <= 1000

    Returns:
        dict
    """

    ids = song_ids if isinstance(song_ids, list) else [song_ids]
    return "/eapi/song/enhance/player/url", {
        "ids": ids,
        "encodeType": str(encodeType),
        "br": str(bitrate),
    }


@EapiCryptoRequest
def GetTrackDownloadURL(song_ids: list, bitrate=320000, encodeType="aac"):
    """PC 端 - 以计费下载方式，获取资源URL

    Args:
        song_ids (list): 歌曲 ID
        bitrate (int, optional): 比特率 (96k SQ 320k HQ 320k+ Lossless/SQ). Defaults to 320000
        encodeType (str, optional) Defaults to `aac`. 使用高 bitrate 值时无视该选项

    Notes:
        song_ids 项目数应 <= 1000

    Returns:
        dict
    """

    ids = song_ids if isinstance(song_ids, list) else [song_ids]
    return "/eapi/song/enhance/download/url", {
        "ids": ids,
        "encodeType": "aac",
        "br": str(bitrate),
    }


@WeapiCryptoRequest
def GetTrackLyrics(song_id: str, lv=-1, tv=-1, rv=-1):
    """网页端 - 获取歌曲歌词

    Args:
        song_id (str): 歌曲 ID
        lv (int, optional): 歌词版本. Defaults to -1.
        tv (int, optional): 翻译版本. Defaults to -1.
        rv (int, optional): 罗马音版本. Defaults to -1.

    `版本` 参数为 `-1` 时均会回退至最新版本

    Returns:
        dict
    """
    return "/weapi/song/lyric", {
        "id": str(song_id),
        "lv": str(lv),
        "tv": str(tv),
        "rv": str(rv),
    }


@WeapiCryptoRequest
def GetTrackComments(song_id, offset=0, limit=20, beforeTime=0):
    """网页端 - 获取歌曲评论

    Args:
        album_id (str): 歌曲ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    """
    return "/weapi/v1/resource/comments/R_SO_4_%s" % song_id, {
        "rid": str(song_id),
        "offset": str(offset),
        "total": "true",
        "limit": str(limit),
        "beforeTime": str(beforeTime * 1000),
    }


@EapiEncipered
@EapiCryptoRequest
@LoginRequiredApi
def SetLikeTrack(trackId, like=True, userid=0, e_r=True):
    """PC端 - 收藏歌曲到 `我喜欢的音乐`

    Args:
        trackId (int): 歌曲 ID
        like (bool, optional): 收藏或取消收藏. Defaults to True.
        userid (int, optional): 暂存. Defaults to 0.

    Returns:
        dict
    """
    return "/eapi/song/like", {
        "trackId": str(trackId),
        "userid": str(userid),
        "like": str(like).lower(),
        "e_r": str(e_r).lower(),
    }

DEFAULT_AUDIO_MATCHER_SESSION_ID = RandomString(16)
@WeapiCryptoRequest
def GetMatchTrackByFP(audioFP: str, duration: float, sessionId=DEFAULT_AUDIO_MATCHER_SESSION_ID):
    """移动端（Chrome 插件） - 听歌识曲
    
    Args:
        audioFP (str): Base64 编码的 AFP. 可参考 https://github.com/mos9527/ncm-afp
        duration (float): FP 时长
        sessionId (str, optional): 未知作用. Defaults to DEFAULT_AUDIO_MATCHER_SESSION_ID.
    """    
    return "/weapi/music/audio/match", {
        'algorithmCode': 'shazam_v2', # 貌似有三种？
        'sessionId': sessionId,
        'duration' : float(duration),   
        'from': 'recognize-song',
        'times':'1',
        'decrypt':'1',
        'rawdata': audioFP
    }


# -*- coding: utf-8 -*-
"""Miscellaneous implementations of some of netease's security algorithms
Implemented:
- `abroad` parameter decryption
"""
import ctypes,urllib.parse

# region secrets
WEAPI_ABROAD_SBOX = (
    82   ,9    ,106  ,-43  ,48   ,54   ,-91  ,56   ,
    -65  ,64   ,-93  ,-98  ,-127 ,-13  ,-41  ,-5   ,
    124  ,-29  ,57   ,-126 ,-101 ,47   ,-1   ,-121 ,
    52   ,-114 ,67   ,68   ,-60  ,-34  ,-23  ,-53  ,
    84   ,123  ,-108 ,50   ,-90  ,-62  ,35   ,61   ,
    -18  ,76   ,-107 ,11   ,66   ,-6   ,-61  ,78   ,
    8    ,46   ,-95  ,102  ,40   ,-39  ,36   ,-78  ,
    118  ,91   ,-94  ,73   ,109  ,-117 ,-47  ,37   ,
    114  ,-8   ,-10  ,100  ,-122 ,104  ,-104 ,22   ,
    -44  ,-92  ,92   ,-52  ,93   ,101  ,-74  ,-110 ,
    108  ,112  ,72   ,80   ,-3   ,-19  ,-71  ,-38  ,
    94   ,21   ,70   ,87   ,-89  ,-115 ,-99  ,-124 ,
    -112 ,-40  ,-85  ,0    ,-116 ,-68  ,-45  ,10   ,
    -9   ,-28  ,88   ,5    ,-72  ,-77  ,69   ,6    ,
    -48  ,44   ,30   ,-113 ,-54  ,63   ,15   ,2    ,
    -63  ,-81  ,-67  ,3    ,1    ,19   ,-118 ,107  ,
    58   ,-111 ,17   ,65   ,79   ,103  ,-36  ,-22  ,
    -105 ,-14  ,-49  ,-50  ,-16  ,-76  ,-26  ,115  ,
    -106 ,-84  ,116  ,34   ,-25  ,-83  ,53   ,-123 ,
    -30  ,-7   ,55   ,-24  ,28   ,117  ,-33  ,110  ,
    71   ,-15  ,26   ,113  ,29   ,41   ,-59  ,-119 ,
    111  ,-73  ,98   ,14   ,-86  ,24   ,-66  ,27   ,
    -4   ,86   ,62   ,75   ,-58  ,-46  ,121  ,32   ,
    -102 ,-37  ,-64  ,-2   ,120  ,-51  ,90   ,-12  ,
    31   ,-35  ,-88  ,51   ,-120 ,7    ,-57  ,49   ,
    -79  ,18   ,16   ,89   ,39   ,-128 ,-20  ,95   ,
    96   ,81   ,127  ,-87  ,25   ,-75  ,74   ,13   ,
    45   ,-27  ,122  ,-97  ,-109 ,-55  ,-100 ,-17  ,
    -96  ,-32  ,59   ,77   ,-82  ,42   ,-11  ,-80  ,
    -56  ,-21  ,-69  ,60   ,-125 ,83   ,-103 ,97   ,
    23   ,43   ,4    ,126  ,-70  ,119  ,-42  ,38   ,
    -31  ,105  ,20   ,99   ,85   ,33   ,12   ,125
)
WEAPI_ABROAD_KEY = 'fuck~#$%^&*(458' # ookay...,
WEAPI_ABROAD_IV = ([ord(s) for s in WEAPI_ABROAD_KEY] * 5)[:64]
# 'abroad' parameter key & CBC IV
# endregion

# region functions to mimic the defficiencies of JS bit operations
def jint(v):return ctypes.c_int(v).value
# Bit shifting only takes care of LS 5 bits
def jls(v,bs):
    return jint(v << (bs & 0x1F))
def jrs(v,bs):
    return jint(v >> (bs & 0x1F))
# some bitwise operations are only done on the least significant 32 bits
def jmask(v):
    return v & 0xFFFFFFFF    
def jxor(v1,v2): return jint( jmask(v1) ^ jmask(v2) )
# endregion

# region common practices
def string_to_charcodes(e):    
    # String to charcode array
    return [ord(char) for char in e] if isinstance(e,str) else e
def char_to_hex(e):
    # Char to hexdecimal [00,FF] string
    c = "0123456789abcdef"
    return c[e >> 4 & 15] + c[e & 15]
def to_hex_string(e):
    # int array to hex string
    return ''.join([char_to_hex(a) for a in e])
def cast_to_signed(e):
    # casts any integer value to signed char
    if e < -128:
        return cast_to_signed(e + 256)
    if (e >= -128 and e <= 127):
        return e
    if (e > 127):
        return cast_to_signed(e - 256)  
def cast_to_multi_signed(b):
    # casts int32 to multiple signed chars
    return [cast_to_signed(b >> 24 & 255),cast_to_signed(b >> 16 & 255),cast_to_signed(b >> 8 & 255),cast_to_signed(b & 255)]
def wm_hex_to_ints_hb(e):
    # Hexstring to integer array
    c = []
    for r in range(0,len(e),2):
        h = jls(int(e[r], 16),4)
        m = int(e[r + 1], 16)
        c.append(cast_to_signed(h + m))
    return c
# endregion 

# region core.js security
def c_quote_int_as_hex(a):
    return '%' + '%'.join([char_to_hex(i) for i in a])
def c_signed_xor(v1,v2): 
    return jxor(cast_to_signed(v1),cast_to_signed(v2))
def c_apply_sbox(src,map):        
    return [map[(i >> 4 & 15) * 16 + (i & 15)] for i in src]
def c_decrypt_abroad_message(hexstring):
    # Decrypts 'result' parameter when 'abroad' flag is true in responses
    source = wm_hex_to_ints_hb(hexstring)    
    result = []
    boxA = WEAPI_ABROAD_IV
    for i in range(0,len(source),64):
        box = source[i:i+64]
        boxB = c_apply_sbox(c_apply_sbox(box,WEAPI_ABROAD_SBOX),WEAPI_ABROAD_SBOX)
        boxC = [c_signed_xor(boxB[i],boxA[i]) for i in range(0,64)]    
        boxD = [cast_to_signed(boxC[i] + cast_to_signed(-boxA[i])) for i in range(0,64)]    
        boxE = [c_signed_xor(boxD[i],WEAPI_ABROAD_IV[i]) for i in range(0,64)]                
        boxA = source[i:i+64]
        result += boxE    
    # last 4 bytes contains length, don't need them for now as we can just strip them away
    result = c_quote_int_as_hex(result[:-4])    
    result = urllib.parse.unquote(result).strip('\x00')  
    return result
# endregion


# -*- coding: utf-8 -*-
"""我的音乐云盘 - Cloud APIs"""
import json
from . import WeapiCryptoRequest, LoginRequiredApi, EapiCryptoRequest, GetCurrentSession

BUCKET = "jd-musicrep-privatecloud-audio-public"


@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveInfo(limit=30, offset=0):
    """PC端 - 获取个人云盘内容

    Args:
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    """
    return "/weapi/v1/cloud/get", {limit: str(limit), offset: str(offset)}


@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveItemInfo(song_ids: list):
    """PC端 - 获取个人云盘项目详情

    Args:
        song_ids (list): 云盘项目 ID

    Returns:
        dict
    """
    ids = song_ids if isinstance(song_ids, list) else [song_ids]
    return "/weapi/v1/cloud/get/byids", {"songIds": ids}


@EapiCryptoRequest
def GetNosToken(
    filename,
    md5,
    fileSize,
    ext,
    ftype="audio",
    nos_product=3,
    bucket=BUCKET,
    local=False,
):
    """移动端 - 云盘占位

    Args:
        filename (str): 文件名
        md5 (str): 文件 MD5
        fileSize (str): 文件大小
        ext (str): 文件拓展名
        ftype (str, optional): 上传类型. Defaults to 'audio'.
        nos_product (int, optional): APP类型. Defaults to 3.
        bucket (str, optional): 转存bucket. Defaults to 'jd-musicrep-privatecloud-audio-public'.
        local (bool, optional): 未知. Defaults to False.

    Returns:
        dict
    """
    return "/eapi/nos/token/alloc", {
        "type": str(ftype),
        "nos_product": str(nos_product),
        "md5": str(md5),
        "local": str(local).lower(),
        "filename": str(filename),
        "fileSize": str(fileSize),
        "ext": str(ext),
        "bucket": str(bucket),
    }


def SetUploadObject(
    stream, md5, fileSize, objectKey, token, offset=0, compete=True, bucket=BUCKET
):
    """移动端 - 上传内容

    Args:
        stream : bytes / File 等数据体 .e.g open('file.mp3')
        md5 : 数据体哈希
        objectKey : GetNosToken 获得
        token : GetNosToken 获得
        offset (int, optional): 续传起点. Defaults to 0.
        compete (bool, optional): 文件是否被全部上传. Defaults to True.

    Returns:
        dict
    """
    r = GetCurrentSession().post(
        "http://45.127.129.8/%s/" % bucket + objectKey.replace("/", "%2F"),
        data=stream,
        params={"version": "1.0", "offset": offset, "complete": str(compete).lower()},
        headers={
            "x-nos-token": token,
            "Content-MD5": md5,
            "Content-Type": "cloudmusic",
            "Content-Length": str(fileSize),
        },
    )
    return json.loads(r.text)


@EapiCryptoRequest
def GetCheckCloudUpload(md5, ext="", length=0, bitrate=0, songId=0, version=1):
    """移动端 - 检查云盘资源

    Args:
        md5 (str): 资源MD5哈希
        ext (str, optional): 文件拓展名. Defaults to ''.
        length (int, optional): 文件大小. Defaults to 0.
        bitrate (int, optional): 音频 - 比特率. Defaults to 0.
        songId (int, optional): 云盘资源ID. Defaults to 0 表示新资源.
        version (int, optional): 上传版本. Defaults to 1.

    Returns:
        dict
    """
    return "/eapi/cloud/upload/check", {
        "songId": str(songId),
        "version": str(version),
        "md5": str(md5),
        "length": str(length),
        "ext": str(ext),
        "bitrate": str(bitrate),
    }


@EapiCryptoRequest
def SetUploadCloudInfo(
    resourceId, songid, md5, filename, song=".", artist=".", album=".", bitrate=128
):
    """移动端 - 云盘资源提交

    注：
        - MD5 对应文件需已被 SetUploadObject 上传
        - song 项不得包含字符 .和/

    Args:
        resourceId (str): GetNosToken 获得
        songid (str): GetCheckCloudUpload 获得
        md5 (str): 文件MD5哈希
        filename (str): 文件名
        song (str, optional): 歌名 / 标题. Defaults to ''.
        artist (str, optional): 艺术家名. Defaults to ''.
        album (str, optional): 专辑名. Defaults to ''.
        bitrate (int, optional): 音频 - 比特率. Defaults to 0.

    WIP - 封面ID,歌词ID 等

    Returns:
        dict
    """
    return "/eapi/upload/cloud/info/v2", {
        "resourceId": str(resourceId),
        "songid": str(songid),
        "md5": str(md5),
        "filename": str(filename),
        "song": str(song),
        "artist": str(artist),
        "album": str(album),
        "bitrate": bitrate,
    }


@EapiCryptoRequest
def SetPublishCloudResource(songid):
    """移动端 - 云盘资源发布

    Args:
        songid (str): 来自 SetUploadCloudInfo

    Returns:
        SetUploadCloudInfo
    """
    return "/eapi/cloud/pub/v2", {
        "songid": str(songid),
    }


@LoginRequiredApi
def SetRectifySongId(oldSongId, newSongId):
    """移动端 - 歌曲纠偏

    Args:
        oldSongId : 欲纠偏的源歌曲ID
        newSongId : 欲纠偏的目标歌曲ID

    Returns:
        dict
    """
    return (
        GetCurrentSession()
        .get(
            "/api/cloud/user/song/match",
            params={"songId": str(oldSongId), "adjustSongId": str(newSongId)},
        )
        .json()
    )



# -*- coding: utf-8 -*-
"""Helper utilites to aid operations w/ API responses"""
from threading import Lock
import time,logging
truncate_length = 64
logger = logging.getLogger("pyncm.helper")

def SubstituteWithFullwidth(string,sub=set('\x00\\/:<>|?*')):
    return "".join([c if not c in sub else chr(ord(c) + 0xFEE0) for c in string])

def Default(default=None):
    def preWrapper(func):
        @property
        def wrapper(*a, **k):
            try:
                return func(*a, **k)
            except Exception as e:
                logger.warn("Failed to get attribute %s : %s" % (func.__name__,e))
                return default
        return wrapper
    return preWrapper

class IDCahceHelper:
    """Generic cache for pesudo-static ID-based dicts"""
    _cache = dict()
    def __new__(cls,item_id,*args):
        if not item_id in IDCahceHelper._cache:            
            IDCahceHelper._cache[item_id] = super().__new__(cls)            
        return IDCahceHelper._cache[item_id]

    def __init__(self,item_id,factory_func) -> None:
        # __init__ is always called after __new__
        if hasattr(self,'_lock'):
            with self._lock: # Ensure update's finished
                return
        # We only want to initialze once
        self._item_id = item_id
        self._factory_func = factory_func
        # Automatically update the dict once we finished loading
        self._lock = Lock()                
        # In case of race conditions when our info is still updating...
        return self.refresh()
    def refresh(self):
        with self._lock:
            self.data = self._factory_func(self._item_id)

class AlbumHelper(IDCahceHelper):
    def __init__(self, item_id):        
        from pyncm.apis.album import GetAlbumInfo
        super().__init__(item_id, GetAlbumInfo)
    
    def refresh(self):        
        logger.debug('Caching album info %s' % self._item_id)
        return super().refresh()

    @Default()
    def AlbumName(self):
        """专辑名"""        
        return self.data["album"]["name"]

    @Default()
    def AlbumAliases(self):
        """专辑别名"""
        return self.data["album"]["alias"]

    @Default()
    def AlbumCompany(self):
        """专辑发行"""
        return self.data["album"]["company"]

    @Default()
    def AlbumBreifDescription(self):
        """专辑概述"""
        return self.data["album"]["breifDesc"]

    @Default()
    def AlbumDescription(self):
        """专辑说明"""
        return self.data["album"]["description"]

    @Default()
    def AlbumPublishTime(self):
        """专辑发布年份"""
        epoch = (
            self.data["album"]["publishTime"] / 1000
        )
        return time.gmtime(epoch).tm_year

    @Default()
    def AlbumSongCount(self):
        """专辑歌曲数"""
        return self.data["album"]["size"]

    @Default()
    def AlbumArtists(self):
        """专辑艺术家"""
        return [_ar["name"] for _ar in self.data["album"]["artists"]]
    
class TrackHelper:
    """Helper class for handling generic track objects"""

    def __init__(self, track_dict) -> None:
        self.__dict__.update({"data":track_dict})

    @property
    def Album(self) -> AlbumHelper:
        """专辑对象，会有更多歌曲元数据"""
        return AlbumHelper(self.data["al"]["id"])        
    
    @Default()
    def ID(self):
        """网易云音乐 ID"""
        return self.data["id"]

    @Default()
    def TrackPublishTime(self):
        """歌曲发布年份"""
        epoch = (
            self.data["publishTime"] / 1000
        )  # stored as unix timestamp though only the year was ever useful
        return time.gmtime(epoch).tm_year

    @Default()
    def TrackNumber(self):
        """歌曲编号 （于专辑）"""
        return self.data["no"]

    @Default(default="Unknown")
    def TrackName(self):
        """歌曲名"""
        assert self.data["name"] != None
        return self.data["name"]

    @Default(default=[])
    def TrackAliases(self):
        """歌曲别名"""
        return self.data["alia"]

    @Default(default="Unknown")
    def AlbumName(self):
        """专辑名"""
        if self.data["al"]["id"]:
            return self.data["al"]["name"]
        else:
            return self.data["pc"]["alb"]

    @Default(
        default="https://p1.music.126.net/UeTuwE7pvjBpypWLudqukA==/3132508627578625.jpg"
    )
    def AlbumCover(self):
        """专辑封面"""
        al = self.data["al"] if "al" in self.data else self.data["album"]
        if al["id"]:
            return al["picUrl"]
        else:
            return "https://music.163.com/api/img/blur/" + self.data["pc"]["cid"]
            # source:PC version's core.js

    @Default(default=["Various Artists"])
    def Artists(self):
        """艺术家名 List"""
        ar = self.data["ar"] if "ar" in self.data else self.data["artists"]
        ret = [_ar["name"] for _ar in ar]
        if not ret.count(None):
            return ret
        else:
            return [self.data["pc"]["ar"]]  # for NCM cloud-drive stored audio

    @Default(default="null")
    def CD(self):
        """专辑 CD"""
        return self.data["cd"]

    @Default()
    def Title(self):
        """保存名 [曲名] - [艺术家名 1,2...,n]"""
        return f'{self.TrackName} - {",".join(self.Artists)}'

    @Default()
    def SanitizedTitle(self):
        """保证兼容大多数文件系统的保存名，
        文件名中若有 Windows 非法字符的存在，这些字符将会被替换为全角版本。
        """

        return f'{SubstituteWithFullwidth(self.TrackName)} - {SubstituteWithFullwidth(",".join(self.Artists))}'



import setuptools, pyncm

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyncm",
    version=pyncm.__version__,
    author="greats3an",
    author_email="greats3an@gmail.com",
    description="NeteaseCloudMusic APIs for Python 3.x 适用于 Python 3 的网易云音乐 API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/greats3an/pyncm",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests"],
    entry_points={"console_scripts": ["pyncm=pyncm.__main__:__main__"]},
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-
"""私享云FM - Cloud FM APIs

`zone`即分区，已知的分区有
- `CLASSICAL` 古典
"""

from .. import EapiCryptoRequest


@EapiCryptoRequest
def GetFmZoneInfo(limit=3, zone="CLASSICAL", e_r=True):
    """获取电台内容

    Args:
        limit (int, optional): 获取数目. Defaults to 3.
        zone (str, optional): 分区. Defaults to "CLASSICAL".
        e_r (bool, optional): [未知]. Defaults to True.

    Returns:
        dict
    """
    return "/eapi/zone/fm/get", {
        "limit": str(limit),
        "zone": zone,
        "e_r": str(e_r).lower(),
    }


@EapiCryptoRequest
def SetSkipFmTrack(songId, zone="CLASSICAL", e_r=True):
    """跳过电台歌曲

    Args:
        songId (int): 歌曲ID
        zone (str, optional): 分区. Defaults to "CLASSICAL".
        e_r (bool, optional): [未知]. Defaults to True.

    Returns:
        dict
    """
    return "/eapi/zone/fm/skip", {
        "songId": str(songId),
        "zone": zone,
        "e_r": str(e_r).lower(),
        "alg": "CLSalternate",
        "time": "0",
    }



# -*- coding: utf-8 -*-
"""Another lrc file parser"""
import re
from collections import defaultdict

def LrcProperty(tagname):
    def wrapper(func):
        @property
        def _wrapper(self):
            if not hasattr(self, tagname):
                return Exception(tagname)
            try:
                return getattr(self, tagname)
            except:
                return Exception(tagname)

        @_wrapper.setter
        def _wrapper(self, v):
            setattr(self, tagname, v)

        return _wrapper

    return wrapper


class LrcRegexes:
    LIDTag_ = re.compile(r"(?<=\[)[^\[\]]*(?=\])")
    LIDTag_Type = re.compile(r"[a-z]{2,}(?=:)")
    LIDTag_Content = re.compile(r"(?<=[a-z]{2}:).*")
    LLyrics_ = re.compile(r"[^\[\]]*$")
    LBrackets = re.compile(r"(?<=\[).*(?=\])")
    LTimestamp = re.compile(r"\d*[\.,:]\d*[\.,:]\d*")


# region Static methods
# Timestamp parsers
def stamp2tag(timestamp):
    mm = int(timestamp / 60)
    ss = int((timestamp - mm * 60))
    xx = int(
        (timestamp - mm * 60 - ss) * 100
    )  # We'd use standard 100th of a second here
    mm, ss, xx = str(mm).rjust(2, "0"), str(ss).rjust(2, "0"), str(xx).rjust(2, "0")
    return f"{mm}:{ss}.{xx}"


def tag2stamp(IDTag):
    IDTag = "".join(LrcRegexes.LTimestamp.findall(IDTag))
    if not IDTag:
        return None
    div = IDTag.split(":")
    if len(div) == 2:
        mm, ss = div
        ss, xx = ss.split(".")
    else:
        mm, ss, xx = div
    timestamp = (
        int(mm) * 60 + int(ss) + int(xx) * (0.1 ** len(xx))
    )  # <- workaround for NE's bizarre tag format
    return timestamp


# endregion


class LrcParser:
    """Parses lrc into mutable dict-like objects"""

    # region Properties
    @LrcProperty("ar")
    def Artist(self):
        pass

    @LrcProperty("al")
    def Album(self):
        pass

    @LrcProperty("ti")
    def Title(self):
        pass

    @LrcProperty("au")
    def Author(self):
        pass

    @LrcProperty("length")
    def Length(self):
        pass

    @LrcProperty("by")
    def LRCAuthor(self):
        pass

    @LrcProperty("offset")
    def Offset(self):
        pass

    @LrcProperty("re")
    def Program(self):
        pass

    @LrcProperty("ve")
    def ProgramVersion(self):
        pass

    # endregion

    def __init__(self, lrc=""):
        """Takes lyrics in `LRC` format,then provides lyrics based on timestamps"""
        # Parsing lrc,line by line
        def EnmurateAttributes():
            for m in dir(self):
                if any(
                    f in m
                    for f in [
                        "__",
                        "Add",
                        "Load",
                        "Clear",
                        "Find",
                        "Update",
                        "Dump",
                        "lyrics",
                    ]
                ):
                    continue
                yield (m, str(getattr(self, m)))

        self.lrcAttributes = list(EnmurateAttributes())
        # This function will only work when no attributes are defined
        self.lyrics = defaultdict(list)
        self._lyrics_sorted = (None, None)
        # Achieve some sort of LRU for lyrics_sorted; Stores ([ID],[sorted dict])
        if not lrc:
            # empty input,we are creating lyrics then
            return
        else:
            self.LoadLrc(lrc)

    @property
    def lyrics_sorted(self):
        """Returns sorted version of the lyrics"""
        lastID, lastDict = self._lyrics_sorted
        if lastID != id(self.lyrics):
            self._lyrics_sorted = (
                id(self.lyrics),
                defaultdict(list, sorted(self.lyrics.items())),
            )
            return self.lyrics_sorted
        else:
            return lastDict

    def LoadLrc(self, lrc):
        """Loads a LRC formmated lryics file"""
        for line in lrc.split("\n"):
            IDTag = LrcRegexes.LIDTag_.findall(line)
            IDTagType = "".join(LrcRegexes.LIDTag_Type.findall("".join(IDTag)))
            IDTagContent = "".join(LrcRegexes.LIDTag_Content.findall("".join(IDTag)))
            Lyrics = "".join(LrcRegexes.LLyrics_.findall(line))
            if IDTagType:
                # Tag's type is set,write as class attribute
                setattr(self, IDTagType, IDTagContent)
            elif IDTag:
                # Tag's type is not set but we got the values,treat as lyrics
                # We'll use the timestamp (into seconds) as lyrics' keys,as hashtables can handle that
                try:
                    for _IDTag in IDTag:
                        # Some LRC lyrics would pile a bunch of timestamps in on a line,like
                        #   [00:01.12][00:08.12]Yeah
                        timestamp = tag2stamp(_IDTag)
                        if timestamp is not None and timestamp >= 0:
                            if not isinstance(self.Offset, Exception):
                                timestamp += float(self.Offset)
                            if Lyrics:
                                self.lyrics[timestamp].append(
                                    (_IDTag, Lyrics)
                                )  # Ignore empty lines
                except:
                    pass
        self.lyrics = self.lyrics_sorted  # sort stuff once loaded

    def AddLyrics(self, timestamp, value):
        # Add 1 or multiple line(s) of lyrics with the same timestamp in seconds
        if not isinstance(value, list):
            value = [value]
        for v in value:
            self.lyrics[timestamp].append((stamp2tag(timestamp), v))
        self.lyrics = (
            self.lyrics_sorted
        )  # this isn't too much for previously sorted items though

    def ClearLyrics(self):
        # Clears the lyrics buffer
        self.lyrics.clear()

    def UpdateLyrics(self, iterable, timestamp_function, lyrics_function):
        """This function takes an iterable,a timestamp function,and a lyrics function

        And for every item in the iterable,timestamp and lyrics will be fetch via the said functions,then get added to our buffer
        """
        for line in iterable:
            timestamp = timestamp_function(line)
            lyrics = lyrics_function(line)
            self.AddLyrics(timestamp, lyrics)

    def DumpLyrics(self, delimiter="\t"):
        """Format current lyrics buffer then spits out a LRC formatted string"""
        lrc = ""
        # Adding tags
        for prop, attr in self.lrcAttributes:
            value = getattr(self, prop)
            if not isinstance(value, Exception):  # If such value do exist
                lrc += f"[{attr}:{value}]" + "\n"
        # Adding lyrics
        for timestamp, lyrics in self.lyrics.items():  # write the sorted one
            Lyrics = []
            for tag, _lyric in lyrics:
                Lyrics.append(_lyric)
            lrc += "\n" + f"[{stamp2tag(timestamp)}]{delimiter.join(Lyrics)}"
        return lrc  # Done

    @staticmethod
    def Find(lyrics, timestamp):
        """Finds closest match in our hashable through binary search. Prioritiezs

        ones that comes before the timestamp, Rather than ones with minimum distance

            lyrics  :   From `lrcparser.lyrics_sorted`

            Returns `(timestamp_seconds,lyrics[(timestamp_tag,lyrics)],indexof)`

            Returns None if nothing is found
        """
        # TODO : Can we make it O(loglog N)?
        def search(val, src: list, l, r):
            """Binary serach"""
            pivot = (l + r) >> 1
            if l >= r or pivot == l or pivot == r:
                return pivot
            in_between = (pivot <= 0 or (src[pivot - 1] <= val)) and val <= src[pivot]
            if in_between:
                if src[pivot] == val:
                    return pivot
                elif src[pivot - 1] == val:
                    return pivot - 1
                else:
                    return pivot
            return (
                search(val, src, pivot, r)
                if src[pivot] < val
                else search(val, src, l, pivot)
            )

        timestamps = list(lyrics.keys())
        index = search(timestamp, timestamps, 0, len(lyrics))
        return (timestamps[index], lyrics[timestamps[index]], index)
